module.exports = function() {

    return this
        .opt()
            .name('opt')
            .long('opt')
            .end()
        .arg()
            .name('arg1')
            .end()
        .arg()
            .name('arg2')
            .arr()
            .end()
        .act(function(opts, args) {
            return { opts: opts, args: args };
        });

};
